open Atoms ;;
open Terms ;;
open Substitutions ;;
open Multiset ;;
open Border ;;

(*
 *
 * Dans ce fichier, j'utilise le calcul des frontières
 * que nous avons fait dans border.ml.
 * Ainsi, pour chaque terme t en input de P, on calcule d'abord 
 * toutes les frontières 
 * (ce qui exige de descendre jusqu'en bas du terme)
 * puis, pour chaque frontière, on calcule les termes aplatis
 * dans P et dans Q, ce qui exige de descendre à nouveau
 * jusqu'au niveau de la frontière dans P.
 * Cet algorithme est quadratique 
 * alors qu'il existe un algorithme linéaire
 * (qui consiste à tout faire d'un coup).
 * Comme je n'arrive pas à coder proprement l'algorithme linéaire,
 * j'ai préféré utiliser l'algorithme quadratique.
 *
 * Les principales difficultés avec l'algorithme linéaires sont:
   1 - On doit traiter le terme côté P
   et celui côté Q en même temps.
   2 - Il n'est pas facile de savoir si l'on a déjà traité
   une frontière donnée.
 *)

let to_ik a b =
  match Array.length a = Array.length b with
  | true -> Array.init (Array.length a) (fun i -> Ik(a.(i),b.(i)))
  | false -> failwith("This function should not be called in this case.")
;;

let array_from t fv =
    Array.init (Array.length t) (fun n -> x (fv + n))
;;


let flat a =
  let rec calcule res = function
    | n when n = Array.length a -> res
    | n -> calcule (res + Array.length a.(n)) (n+1)
  in let k = calcule 0 0
  in let rec get_some = function
    | i when i  = Array.length a -> raise Not_found
    | i when Array.length a.(i) = 0 -> get_some (i+1)
    | i -> a.(i).(0)

  in let arr =
   match k with
   | 0 -> [||]
   | j ->  Array.make j (get_some 0)
  in let rec copy c = function
    |(i,_) when i = Array.length a -> assert(c=k)
    |(i,j) when j = Array.length a.(i) -> copy c (i+1,0)
    |(i,j) -> arr.(c) <- a.(i).(j) ; copy (c+1) (i,j+1)
  in copy 0 (0,0) ; arr
;;


let rec cut_type_p fv tp border =
  match border with
  | At() -> [|At(Var(fv),tp)|],fv+1
  |Omega  -> assert(false)
  | t -> cut_type_array_p fv (sons tp) (sons t)

and cut_type_array_p fv aTp aBorder =
  let res = Array.make (Array.length aTp) ([|At(Var(0),aTp.(0))|])
  in let rec cut fv = function
    | i when i = Array.length aTp -> fv
    | i -> 
        let arr,fv2 =  cut_type_p fv aTp.(i) aBorder.(i) 
        in res.(i) <- arr ; cut fv2 (i+1)
  in let fv2 = cut fv 0
  in flat res,fv2
;;

let rec cut_var_q fv border =
  match border with
  |At() -> [|At(Var(fv))|],fv+1
  |Omega -> assert(false)
  | t -> cut_var_array_q fv (sons t)

and cut_var_array_q fv aBorder =
  let res = Array.make (Array.length aBorder) ([|At(Var(0))|])
  in let rec cut fv = function
    | i when i = Array.length aBorder -> fv
    | i -> 
        let arr,fv2 = cut_var_q fv aBorder.(i) 
        in res.(i) <- arr ; cut fv2 (i+1)
  in let fv2 = cut fv 0
  in flat res,fv2
;;


let rebuild arr border =
  let rec aux i = function
    | At() -> arr.(i),i+1
    | Omega -> failwith("This function should not be called on Omega.")
    | Senc(a,b) -> 
        let t1,j1 = aux i a in
        let t2,j2 = aux j1 b in
        Senc(t1,t2),j2
    | Aenc(a,b) -> 
        let t1,j1 = aux i a in
        let t2,j2 = aux j1 b in
        Aenc(t1,t2),j2
    | Sign(a,b) -> 
        let t1,j1 = aux i a in
        let t2,j2 = aux j1 b in
        Sign(t1,t2),j2
    | Pub(a) -> 
        let t1,j1 = aux i a in
        Pub(t1),j1
    | Hash(a) -> 
        let t1,j1 = aux i a in
        Hash(t1),j1
    | Vk(a) -> 
        let t1,j1 = aux i a in
        Vk(t1),j1
    | Tuple(t) -> 
        let t1,i1 = aux_array i t in
        Tuple(t1),i1
  and aux_array i t =
    let t1,i1 = aux i t.(0) in
    let res = Array.make (Array.length t) t1 in
    let rec on_ith j = function
      | n when n = Array.length t ->  j
      | n -> 
          let tn,jn = aux j t.(n) 
          in res.(n) <- tn ; on_ith jn (n+1)
    in let j = on_ith i1 1 in
    res,j
  in let t,j = aux 0 border
  in assert(j = Array.length arr); t

  (* On suppose que l'on a effectué le calcul
   * sur un tableau et une frontière qui vont ensemble,
   * et donc le nombre de trous à remplir dans la frontière
   * correspond au nombre d'éléments dans le tableau. *)
;;

let rec cut_term_p fv t border =
  match t,border with
  |x,At() -> Substitutions.make (),[|x|],fv
  |_,Omega|Omega,_ -> 
      failwith("This function should not be called on Omega.")
  |At(Var(n),tp),y -> 
      let arr,fv1 = cut_type_p fv tp y in
      let tRebuilt = rebuild arr y in
      let sigma = Substitutions.make () in
      let sigma1 = add_value sigma (n,tRebuilt) in
      sigma1,arr,fv1
  
  (* On pourrait regrouper les trois cas suivants
   * avec le cas général. *)
  |Pub(a),Pub(b) -> cut_term_p fv a b
  |Hash(a),Hash(b) -> cut_term_p fv a b
  |Vk(a),Vk(b) -> cut_term_p fv a b

  (* Cas général *)
  |t,b when same_root t b ->
      let st = sons t in
      let sb = sons b in
      cut_term_array_p fv st sb
  | _ -> failwith("In this function, border should be a border of t.")

and cut_term_array_p fv aT aBorder =
  let res = Array.make (Array.length aT) [||]
  in let rec cut sigma fv = function
    | i when i = Array.length aT -> sigma,fv
    | i -> 
        let aTi = typed_term_map sigma aT.(i) in
        let sigma1,arr,fv2 = cut_term_p fv aTi aBorder.(i) in
        let sigma2 = merge sigma sigma1
        (* La fusion est autorisée car on a instancié d'abord
         * avec sigma. *)
        in res.(i) <- arr ; cut sigma2 fv2 (i+1)
  in let sigma1,fv1 = cut (Substitutions.make ()) fv 0
  in sigma1,flat res,fv1
;;

let rec cut_term_q fv t border =
  match t,border with
  |x,At() -> true,Substitutions.make (),[|x|],fv
  |_,Omega|Omega,_ -> 
      failwith("This function should not be called on Omega.")
  |At(Var(n)),y -> 
      let arr,fv1 = cut_var_q fv y in
      let tRebuilt = rebuild arr y in
      let sigma = Substitutions.make () in
      let sigma1 = add_value sigma (n,tRebuilt) in
      true,sigma1,arr,fv1

  (* Cas q a la même forme que p *)
  |t,b when same_root t b ->
      let st = sons t in
      let sb = sons b in
      cut_term_array_q fv st sb

  (* Cas q n'a pas la même forme que p *)

  |_,b ->
      let sb = sons b in 
      let st = array_from sb fv in
      let fv1 = fv + Array.length sb in
      let _,sigma,arr,fv2 = cut_term_array_q fv1 st sb in
      false,sigma,arr,fv2

and cut_term_array_q fv aT aBorder =
  let res = Array.make (Array.length aT) [||]
  in let rec cut fidele sigma fv = function
    | i when i = Array.length aT -> fidele,sigma,fv
    | i -> 
        let aTi = untyped_term_map sigma aT.(i) in
        let fidele1,sigma1,arr,fv2 = cut_term_q fv aTi aBorder.(i) in
        let sigma2 = merge sigma sigma1 in
        (* La fusion est autorisée car on a instancié d'abord
         * avec sigma. *)
        let fidele2 = fidele1 && fidele in
        res.(i) <- arr ; cut fidele2 sigma2 fv2 (i+1)
  in let fidele1,sigma1,fv1 = cut true (Substitutions.make ()) fv 0
  in fidele1,sigma1,flat res,fv1
;;

let get_all_flats_versions fv ik =
  let get_left = function
    | Ik(a,b) -> a
    | _ -> failwith("This function should only be called on Ik facts.")
  in let get_right = function
    | Ik(a,b) -> b
    | _ -> failwith("This function should only be called on Ik facts.")
  in let ik_P = Array.map get_left ik
  in let ik_Q = Array.map get_right ik
  in let border_list = borders ik_P in
  let flatP = List.map (cut_term_array_p fv ik_P) border_list in
  let flatQ = List.map (cut_term_array_q fv ik_Q) border_list in
  let rec merge_lists res = function
    | [],[] -> res
    | (sigmaP,arrP,fvP)::tp,(fidele,sigmaQ,arrQ,fvQ)::tq ->
        let new_res = 
          (fidele,sigmaP,sigmaQ,to_ik arrP arrQ,max fvP fvQ) :: res
        in merge_lists new_res (tp,tq)
    | _ -> failwith("The lists should have the same length.")
  in merge_lists [] (flatP,flatQ)
;;

