open Atoms ;;
open Terms ;;
open Multiset ;;
open Process ;;


(* Règles pour l'équivalence statique. *)
let static_sym = 
        [
                Adv(5,[|x 0|],[|x 1|]),
                [|Ik(yt 0 (At(Cameleon)),x 1) ; Ik(yt 0 (At(Cameleon)),x 1)|],
                [||],
                [||]
        ]
;;
let static_rules =
        let new_rule =
                (
                Adv(4,[|x 0;x 1|],[|x 2;x 3|]),
                [|Ik(Sign(yt 0 (At(Cameleon)), At(Var( 1 ) ,At(Cameleon))),Sign(x 2,x 3));Ik(Vk(At(Var(1),At(Cameleon))),Vk(x 3))|],
                [||],
                [||]
                )
        in new_rule::static_sym
;;

let types liste =

  let rec insert a = function
    | [] -> [a]
    | x::t when a = x -> x::t
    | x::t -> x::(insert a t)
  in
  let rec signed_subterms l = function
    | Sign(t1,t2) -> signed_subterms (signed_subterms (insert t1 l) t2) t1
    | Tuple(t1) -> Array.fold_left signed_subterms l t1
    | Senc(t1,t2) -> signed_subterms (signed_subterms l t1) t2
    | Aenc(t1,t2) -> signed_subterms (signed_subterms l t1) t2
    (* Inutile de descendre en dessous du hash
     * car il n'est pas inversible. *)
    | Hash(_)     -> l
    | Pub(t)      -> signed_subterms l t
    | Vk(t)       -> signed_subterms l t
    | Omega       -> l
    | At(_)       -> l
  in
  let rec flat l = function
    | Tuple(t) -> Array.fold_left flat l t
    | Sign(t1,t2) -> insert (Sign(t1,t2)) l
    | Senc(t1,t2) -> flat (flat l t1) t2
    | Aenc(t1,t2) -> insert (Aenc(t1,t2)) l 
    | Hash(t1)    -> insert (Hash(t1)) l
    | Pub(t1)     -> insert (Pub(t1))  l
    | Vk(t1)      -> insert (Vk(t1))   l
    | Omega       -> l
    | At(_)       -> l
  in
  let rec types_process res = function
    | PhaseRestriction(_,next) -> types_process res next
    | In(_,_,next) -> types_process res next
    | Out(_,t,next) -> 
        let tp_t = get_type t in
        let res1 = flat res tp_t in
        let signed = signed_subterms [] tp_t in
        let res2 = List.fold_left flat res1 signed in
        types_process res2 next
    | Null -> res

  in let rec aux res = function
    | [] -> res
    | (p,_)::t -> aux (types_process res p) t
  in aux [] liste
;;

let simplify liste =
  let rec insert a = function
    | [] -> [a]
    | x::t when x = a -> x::t
    | x::t -> x::(insert a t)
  in let rec all = function
    | [] -> []
    | a::t -> insert a (all t)
  in all liste
;;

(* l est une liste de terme,terme,atoms,vars
 * t est un terme représentant le symbol de fonction de tête *)

let rec add_symbol1 t l =
        (* Seul le symbole de tête de t compte.
         * Il sert à repérer le symbole qu'on doit ajouter. *)
        match t,l with
        | _,[] -> []
        | Hash(_),(x1,x2,atoms,vars)::l -> 
            ((Hash x1), (Hash x2),atoms,vars)::(add_symbol1 t l)
        | Pub(_),(x1,x2,atoms,vars)::l -> 
            ((Pub x1), (Pub x2),atoms,vars)::(add_symbol1 t l)
        | Vk (_),(x1,x2,atoms,vars)::l -> 
            ((Vk  x1), (Vk  x2),atoms,vars)::(add_symbol1 t l)
        | _ -> raise (Invalid_argument "should be used with arity one function symbols")
;;

(* Problème il faut utiliser des tableaux *)

let add_symbol2 t l1 l2 =
        let rec etape1 res t0 (x1,x2,atoms,vars) l2 =
             match t0,l2 with
             | _,[] -> res

             | Senc(_,_),(y1,y2,a,v)::l -> 
                let natoms = List.rev_append a atoms in
                let nvars  = List.rev_append v vars  in
                etape1 ((Senc(x1,y1),Senc(x2,y2),natoms,nvars)::res) 
                t0 (x1,x2,atoms,vars) l

             | Aenc(_,_),(y1,y2,a,v)::l -> 
                let natoms = List.rev_append a atoms in
                let nvars  = List.rev_append v vars  in
                etape1 ((Aenc(x1,y1),Aenc(x2,y2),natoms,nvars)::res) 
                t0 (x1,x2,atoms,vars) l

             | Sign(_,_),(y1,y2,a,v)::l -> 
                let natoms = List.rev_append a atoms in
                let nvars  = List.rev_append v vars  in
                etape1 ((Sign(x1,y1),Sign(x2,y2),natoms,nvars)::res) 
                t0 (x1,x2,atoms,vars) l

             | _ -> 
                raise 
                (Invalid_argument 
                "should be used with arity two function symbols")

        in let rec etape2 res t0 l1 l2 =
                match l1 with
                | [] -> res
                | x::l -> etape2 (etape1 res t0 x l2) t0 l l2
        in etape2 [] t l1 l2
;;


(* Cette fonction ajoute le tuple sur les contextes
 * donnés dans le tableau l_arr.
 * l_arr est un tableau dont la longueur est
 * la longueur du tuple et chaque case contient
 * tous les contextes attendus à cet endroit. *)

let add_tuple l_arr =
  
  (* Première étape : On procédure récursivement.
   * On suppose que l'on a chacun des débuts de tuples
   * possibles pour les n premières cases,
   * et l'on cherche à obtenir toutes les possibilités jusqu'à
   * la n+1 ème case.
   * Pour ce faire, on considère d'abord une seule
   * des possibilités que l'on a obtenues avant
   * et on lui ajoute ce qu'il faut en fonction de la liste
   * contenue dans la n+1 ème case.
   * Attention : la liste est construite à l'envers. *)

  let rec etape1 res (pre1,pre2,atoms,vars) = function
    |[] -> res
    |(y1,y2,a,v)::l ->
        let natoms = List.rev_append a atoms in
        let nvars  = List.rev_append v vars  in
        etape1 ((y1::pre1,y2::pre2,natoms,nvars)::res)
        (pre1,pre2,atoms,vars) l
  
  (* Deuxième étape : Il s'agit maintenant d'agir sur toutes
   * les combinaisons à l'étape précédente. *)
  in let rec etape2 n res = function
    | [] -> res
    | (pre1,pre2,atoms,vars)::l -> 
        etape2 n (etape1 res (pre1,pre2,atoms,vars) l_arr.(n)) l
  
  (* Troisième étape : Il suffit d'itérer sur toutes les étapes
   * Comme on procède aussi à l'envers,
   * les listes inversées à l'étape 1 sont en fait à l'endroit. *)
  in let rec etape3 res3 = function
    | 0 -> res3
    | n -> etape3 (etape2 (n-1) [] res3) (n-1)

  (* Dernière étape : Il faut transformer les listes en tableaux
   * et mettre le constructeur Tuple dessus. *)

  in let rec etape4 res = function
    | [] -> res
    | (l1,l2,atoms,vars)::l -> 
        let tuple1 = Tuple(Array.of_list l1) in
        let tuple2 = Tuple(Array.of_list l2) in
        etape4 ((tuple1,tuple2,atoms,vars)::res) l

  in etape4 [] (etape3 [] (Array.length l_arr))
;;

let for_one_item ruleNumber t1 t2 atoms vars =
  (* *)
  let rec create_ik_list = function
    | ([],[]) -> []
    | x::lx,y::ly -> (Ik(x,y))::(create_ik_list (lx,ly))
    | _ -> raise (Invalid_argument "The length of the lists differ")
  in let pre = Array.of_list (Ik(t1,t2) :: create_ik_list (atoms,vars))
  in ((Adv(ruleNumber,[||],Array.of_list vars),pre,[||],[||]):rule)
;;


(* Cette fonction agit sur les types du protocole.
 * *)

let create_items t =
  let rec aux fv = function
    (* Si le type est atomique, il n'y a plus rien à faire :
      * on ajoute une variable du bon type pour P et pour Q.
      * *)
    | At(t) ->
       let (new_y:typed_term) = yt fv (At(t)) in
       let new_x = x (fv+1) in
       [new_y,new_x,[new_y],[new_x]],fv+2
    | Omega -> 
       raise (Invalid_argument "Omega is not a type." )
    (* Si le type est composé, 
     * il faut :
       * 1 - récupérer tous les contextes plus petits
       * 2 - ajouter le symbole au-dessus de tous ces contextes
       * 3 - créer de nouvelles variables et donner le résultat *)
    | Hash(t)-> 
       let l,nfv = aux fv t in
       let new_l = add_symbol1 (Hash t) l in
       let new_y = yt nfv t in
       let new_x = x (nfv+1) in
       (((Hash(new_y)):typed_term),new_x,[Hash(new_y)],[new_x])::new_l,
       nfv+1

    | Pub(t)-> 
       let l,nfv = aux fv t in
       let new_l = add_symbol1 (Pub t) l in
       let new_y = yt nfv t in
       let new_x = x (nfv+1) in
        (((Pub(new_y)):typed_term),new_x,[Pub(new_y)],[new_x])::new_l,
        nfv+1

    | Vk(t) ->  
       let l,nfv = aux fv t in
       let new_l = add_symbol1 (Vk t) l in
       let new_y = yt nfv t in
       let new_x = x (nfv+1) in
       (((Vk(new_y)):typed_term),new_x,[Vk(new_y)],[new_x])::new_l,
       nfv+1

    | Senc(t1,t2) ->  
        let l1,fv1 = aux fv  t1 in
        let l2,fv2 = aux fv1 t2 in
        let new_l = add_symbol2 (Senc (t1,t2)) l1 l2 in
        (Senc(yt fv2 t1,yt (fv2+1) t2),x (fv2+2),
        [Senc(yt fv2 t1,yt (fv2+1) t2)],[x (fv2+2)])::new_l,fv2+3

    | Aenc(t1,t2) ->  
        let l1,fv1 = aux fv  t1 in
        let l2,fv2 = aux fv1 t2 in
        let new_l = add_symbol2 (Aenc (t1,t2)) l1 l2 in
        (Aenc(yt fv2 t1,yt (fv2+1) t2),x (fv2+2),
        [Aenc(yt fv2 t1,yt (fv2+1) t2)],[x (fv2+2)])::new_l,fv2+3

    | Tuple(t) ->  
        let l_arr,fv1 = aux_arr fv t in
        let new_l = add_tuple l_arr in
        let y_arr = Array.init (Array.length t) 
                               (fun n -> yt (fv1+n) t.(n) )
        in let fv2 = fv1+Array.length t in
        let new_x = x fv2 in 
        (Tuple(y_arr),new_x,[Tuple(y_arr)],[new_x])::new_l,fv2+1

    | Sign(t1,t2) ->  
        let l1,fv1 = aux fv  t1 in
        let l2,fv2 = aux fv1 t2 in
        let new_l = add_symbol2 (Sign (t1,t2)) l1 l2 in
        (Sign(yt fv2 t1,yt (fv2+1) t2),x (fv2+2),
        [Sign(yt fv2 t1,yt (fv2+1) t2)],[x (fv2+2)])::new_l,fv2+3

  
  and aux_arr fv arr =
    let res = Array.make (Array.length arr) [] in
    let rec help fv = function
      | n when n >= Array.length arr -> res,fv
      | n -> 
          let l,fv1 = aux fv (arr.(n)) in
          begin
            res.(n) <- l ;
            help fv1 (n+1)
          end
    in help fv 0
  in aux 0 t
 (* on aurait pu éliminer une partie de ces règles *)
;;


(*
 * Cette fonction déduit les règles d'équivalence statique.
 * assym nous dit si nous avons des primitives asymétriques
 * proto nous donne le protocole.
 *)

let all_static_rules predefined_rules assym proto =

(* première étape : récupérer les types du protocole *)
  let terms = simplify (types proto)
(* deuxième étape : construire les contextes *)
  in let rec on_all_items res ruleNumber = function
    | [] -> res,ruleNumber
    | (x1,x2,atoms,vars)::t -> 
       on_all_items 
       ((for_one_item ruleNumber x1 x2 atoms vars)::res) 
       (ruleNumber + 1) t
  in let rec aux res ruleNumber = function
    |[] -> res
    |t0::l -> 
    let items,_  = create_items t0 in 
    let res1,rN1 = on_all_items res ruleNumber items in
        aux res1 rN1 l
  in match assym with
    |false -> static_sym
    |true  -> aux static_rules predefined_rules terms
;;


