open Atoms ;;

(* Constructor Terms *)

type 'a term = 
        | At of 'a 
        | Omega 
        (* Omega est la constante <w,w> qui fait la différence entre typé et pseudo-typé. *)
        | Senc  of 'a term * 'a term 
        | Tuple of 'a term array
        | Aenc  of 'a term * 'a term
        | Hash  of 'a term
        | Pub   of 'a term
        | Sign  of 'a term * 'a term
        | Vk of 'a term
;;

let top_arity = function
  | At(_)     -> 0
  | Omega     -> 0
  | Hash(_)   -> 1
  | Pub(_)    -> 1
  | Vk(_)     -> 1
  | Sign(_,_) -> 2
  | Senc(_,_) -> 2
  | Aenc(_,_) -> 2
  | Tuple(t)  -> Array.length t
;;

let sons = function
  | At(_) -> failwith("This function should not be used with data")
  | Omega -> failwith("This function should not be used with data")
  | Senc(a,b) -> [|a;b|]
  | Aenc(a,b) -> [|a;b|]
  | Sign(a,b) -> [|a;b|]
  | Hash(a)   -> [|a|]
  | Pub(a)    -> [|a|]
  | Vk(a)     -> [|a|]
  | Tuple(t)  -> t

let same_root t1 t2 =
  match t1,t2 with
  | At(_),At(_)|Omega,Omega|Senc(_,_),Senc(_,_)
  | Aenc(_,_),Aenc(_,_)|Sign(_,_),Sign(_,_)|Hash(_),Hash(_)
  |Pub(_),Pub(_)|Vk(_),Vk(_) -> true
  |Tuple(t1),Tuple(t2) -> Array.length t1 = Array.length t2
  | _ -> false
;;



(* Quick atoms constructors : *)

let x n   =  At(Var(n)) ;;
let c n   =  At(Cst(n)) ;;
let nce n =  At(Nonce(n)) ;;

(* Print function *)

let to_string_generic at_to_string t =
  let rec to_string_tuple_array a = function
    | n when n < (Array.length a)-1 -> 
        begin
        String.concat "," [main a.(n) ; to_string_tuple_array a (n+1)]
        end
    | n (* when n = Array.length a - 1 *) -> main a.(n) 

  and main = function
        | At(x) -> at_to_string x
        | Omega -> "<w,w>"
        | Senc(a,b) -> String.concat "" ["senc(" ; main a ; "," ; main b ; ")" ]
        | Tuple(a)  -> Printf.sprintf "<%s>" (to_string_tuple_array a 0)
        | Aenc(a,b) -> Printf.sprintf "aenc(%s,%s)" (main a) (main b) 
        | Hash(a)   -> Printf.sprintf "hash(%s)" (main a) 
        | Pub(a)    -> Printf.sprintf "pub(%s)" (main a)
        | Sign(a,b) -> Printf.sprintf "sign(%s,%s)" (main a) (main b) 
        | Vk(a)     -> Printf.sprintf "vk(%s)" (main a)

        in main t
;;

let print_term t = print_string (to_string_generic Atoms.to_string t) ;;

(* Generic functions on terms *)

let is_ground m =
  let rec acc = function
    | a::l -> aux l a
    | [] -> true
  and acc_array l a = function
    | n when n < Array.length a -> acc_array (a.(n)::l) a (n+1)
    | _ -> acc l
  and aux l = function
    | At(x) -> is_ground_atom x
    | Omega -> acc l
    | Pub(x) -> aux l x
    | Vk(x)  -> aux l x
    | Tuple(a) -> acc_array l a 0
    | Senc(a,b) -> acc (a::b::l)
    | Aenc(a,b) -> acc (a::b::l)
    | Sign(a,b) -> acc (a::b::l)
    | Hash(x) -> aux l x
  in aux [] m
;;


let rec max_tuple_size_term n = function
  | Tuple(t) ->
      let m = Array.fold_left max_tuple_size_term n t in
      max m (Array.length t)
  | At(_) -> n
  | Omega -> n
  | Pub(a) -> max_tuple_size_term n a
  | Vk(a)  -> max_tuple_size_term n a
  | Hash(a) -> max_tuple_size_term n a
  | Senc(a,b) -> max_tuple_size_term (max_tuple_size_term n a) b
  | Aenc(a,b) -> max_tuple_size_term (max_tuple_size_term n a) b
  | Sign(a,b) -> max_tuple_size_term (max_tuple_size_term n a) b
;;



(* Boolean functions on terms *)

let is_atomic = function
  | At(_) -> true
  | _ -> false
;;

let is_pub = function
  | Pub(_) -> true
  | _ -> false
;;

let has_atomic_keys t =
  let rec acc = function
    | a::l -> aux l a
    | [] -> true 
  and acc_array l a = function
    | n when n < Array.length a -> acc_array (a.(n)::l) a (n+1)
    | _ -> acc l
  and aux l = function
    |At(_) -> acc l
    |Omega -> acc l
    |Pub(At(_)) -> acc l
    |Vk(At(_)) -> acc l
    |Hash(x) -> aux l x
    |Tuple(a) -> acc_array l a 0
    |Senc(a,At(_)) -> aux l a
    |Aenc(a,Pub(At(_))) -> aux l a
    |Sign(a,At(_)) -> aux l a 
    |_ -> false
  in aux [] t
;;  

(* Types *)

type atom_type = Atomic_type of string | Cameleon ;;

type complete_type = atom_type term ;; 
(* Type Message is only used for analyze rules which may be untyped *)

let type_checks t1 t2 =
  let rec acc = function
    | a::l -> aux l a
    | [] -> true 
  and array_acc l a1 a2 =
    match Array.length a1 = Array.length a2 with
    |true  -> array_acc_inside l a1 a2 0
    |false -> false
  and array_acc_inside l a1 a2 = function
    | n when n < Array.length a1 -> array_acc_inside ((a1.(n),a2.(n))::l) a1 a2 (n+1)
    | _ -> acc l
  and aux l = function
    | At(Cameleon), _  -> acc l
    | _ , At(Cameleon) -> acc l
    | At(x),At(y)   -> (x = y)&&(acc l)
    | Pub(x),Pub(y) -> aux l (x,y)
    | Vk(x), Vk(y)  -> aux l (x,y)
    | Hash(x), Hash(y) -> aux l (x,y)
    | Senc(a,b),Senc(c,d) -> acc ((a,c)::(b,d)::l)
    | Aenc(a,b),Aenc(c,d) -> acc ((a,c)::(b,d)::l)
    | Sign(a,b),Sign(c,d) -> acc ((a,c)::(b,d)::l)
    | Tuple(a), Tuple(b)  -> array_acc l a b
    | _ -> false
  in aux [] (t1,t2)
;;

(* Atoms, and in particular variables, may have a non atomic type : *)

type typed_term = (atom*complete_type) term ;;

let rec untype = function
  | At(x,_) -> At(x)
  | Omega -> Omega
  | Pub(a) -> Pub(untype a)
  | Vk(a)  -> Vk (untype a)
  | Hash(a)  -> Hash (untype a)
  | Sign(a,b) -> Sign(untype a,untype b)
  | Aenc(a,b) -> Aenc(untype a,untype b)
  | Senc(a,b) -> Senc(untype a,untype b)
  | Tuple(a) -> Tuple(Array.map untype a)
;;

let rec get_type = function
  | At(_,t) -> t
  | Omega   -> At(Cameleon) (* Il n'y a pas de contrôle de type sur Omega *)
  | Pub(a) -> Pub(get_type a)
  | Vk(a)  -> Vk (get_type a)
  | Hash(a)  -> Hash (get_type a)
  | Sign(a,b) -> Sign(get_type a, get_type b)
  | Aenc(a,b) -> Aenc(get_type a, get_type b)
  | Senc(a,b) -> Senc(get_type a, get_type b)
  | Tuple(a) -> Tuple(Array.map get_type a)
;;

let max_tuple_size_typed n t =
  let tp = get_type t in
  max_tuple_size_term (max_tuple_size_term n tp) t
;;

let type_matches t1 t2 = type_checks (get_type t1) (get_type t2) ;;

(* Quick terms constructors *)

let xt n str   =  At(Var(n),At(Atomic_type(str))) ;;
let yt n tp    =  At(Var(n),tp) ;;
let ct n str   =  At(Cst(n),At(Atomic_type(str))) ;;
let ncet n str =  At(Nonce(n),At(Atomic_type(str))) ;;
let nce_from_tp n tp =  At(Nonce(n),tp) ;;
let pub n str = Pub(At(Var(n),At(Atomic_type(str))));;
let vk  n str = Vk( At(Var(n),At(Atomic_type(str))));;

(* Print functions for types, terms of types and typed terms *)

let atom_type_to_string = function
  | Atomic_type t -> String.concat "" ["'t_" ; t ] 
  | Cameleon -> "cameleon"
;;

let term_of_type_to_string (t0:atom_type term) = to_string_generic atom_type_to_string t0 ;;

let complete_type_to_string t = term_of_type_to_string t ;;


let typed_term_to_string (t0:typed_term) = to_string_generic (fun (x,t) -> String.concat " : " [Atoms.to_string x ; complete_type_to_string t ])  t0 ;;
let print_typed_term t0 = print_string (typed_term_to_string t0);;

let without_type_to_string (t0:typed_term) = to_string_generic (fun (x,_) -> Atoms.to_string x) t0 ;;
let print_term_without_type t0 = print_string (without_type_to_string t0) ;;
